# Copyright 2014, 2015 Red Hat
#
# Authors: Chris Dent <chdent@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os
import re

import jsonpath_rw
import testtools
from six.moves.urllib import parse as urlparse


class HTTPTestCase(testtools.TestCase):
    """Encapsulate a single HTTP request as a TestCase.

    If the test is a member of a sequence of requests, ensure that prior
    tests are run.

    To keep the test harness happy we need to make sure the setUp and
    tearDown are only run once.
    """

    def setUp(self):
        if not self.has_run:
            super(HTTPTestCase, self).setUp()

    def tearDown(self):
        if not self.has_run:
            super(HTTPTestCase, self).tearDown()
        self.has_run = True

    def test_request(self):
        """Run this request if it has not yet run.

        If there is a prior test in the sequence, run it first.
        """
        if self.has_run:
            return
        if self.prior and not self.prior.has_run:
            self.prior.run()
        self._run_test()

    def _assert_response(self, response, content, status, headers=None,
                         expected=None, json_paths=None):
        """Compare the response with expected data."""

        # Never accept a 500
        if response['status'] == '500':
            raise ServerError(content)

        # Always test status
        self._test_status(status, response['status'])

        if headers:
            self._test_headers(headers, response)

        decoded_output = self._decode_content(response, content)

        # Compare strings in response body
        if expected:
            for expect in expected:
                self.assertIn(expect, decoded_output)

        # Decode body as JSON and compare.
        # NOTE(chdent): This just here for now to see if it is workable.
        if json_paths:
            response_data = json.loads(decoded_output)
            self.json_data = response_data
            for path in json_paths:
                match = self._extract_json_path_value(response_data, path)
                self.assertEqual(json_paths[path], match)

    def _decode_content(self, response, content):
        """Decode content to a proper string."""
        content_type = response.get('content-type',
                                    'application/binary').lower()
        if ';' in content_type:
            content_type, charset = (attr.strip() for attr in
                                     content_type.split(';'))
            charset = charset.split('=')[1].strip()
        else:
            charset = 'utf-8'

        if self._not_binary(content_type):
            return content.decode(charset)
        else:
            return content

    @staticmethod
    def _extract_json_path_value(data, path):
        """Extract the value at JSON Path path from the data.

        The input data is a Python datastructre, not a JSON string.
        """
        path_expr = jsonpath_rw.parse(path)
        matches = [match.value for match in path_expr.find(data)]
        return matches[0]

    def _load_data_file(self, filename):
        """Read a file from the current test directory."""
        path = os.path.join(self.test_directory, os.path.basename(filename))
        with open(path, mode='rb') as data_file:
            return data_file.read()

    @staticmethod
    def _not_binary(content_type):
        """Decide if something is content we'd like to treat as a string."""
        return (content_type.startswith('text/')
                or content_type.endswith('+xml')
                or content_type.endswith('+json')
                or content_type == 'application/javascript'
                or content_type == 'application/json')

    def _parse_url(self, url, ssl=False):
        """Create a url from test data.

        If provided with a full URL, just return that. If SSL is requested
        set the scheme appropriately.

        Scheme and netloc are saved for later use in comparisons.
        """
        parsed_url = urlparse.urlsplit(url)
        url_scheme = parsed_url[0]
        scheme = 'http'
        netloc = self.host

        if not url_scheme:
            if self.port:
                netloc = '%s:%s' % (self.host, self.port)
            if ssl:
                scheme = 'https'
            full_url = urlparse.urlunsplit((scheme, netloc, parsed_url[2],
                                            parsed_url[3], ''))
            self.scheme = scheme
            self.netloc = netloc
        else:
            full_url = url
            self.scheme = url_scheme
            self.netloc = parsed_url[1]

        return full_url

    def _replacer(self, match):
        """Replace a regex match with the value of a JSON Path."""
        path = match.group(1)
        return str(self._extract_json_path_value(self.prior.json_data, path))

    def _replace_response_values(self, template):
        """Replace a JSON Path in a template string with its value."""
        return re.sub(r"\$RESPONSE\['([^']+)'\]", self._replacer, template)

    def _run_test(self):
        """Make an HTTP request and compare the response with expectations."""
        test = self.test_data
        http = self.http
        base_url = test['url']

        # Reset follow_redirects with every go.
        http.follow_redirects = False
        if test['redirects']:
            http.follow_redirects = True

        if '$LOCATION' in base_url:
            # Let AttributeError raise
            base_url = base_url.replace('$LOCATION', self.prior.location)

        if '$RESPONSE' in base_url:
            base_url = self._replace_response_values(base_url)

        full_url = self._parse_url(base_url, test['ssl'])
        method = test['method'].upper()
        headers = test['request_headers']

        body = test['data']
        if body:
            body, is_str = self._test_data_to_string(body)
            if self._not_binary(headers['content-type']):
                if is_str:
                    try:
                        body = str(body, 'UTF-8')
                    except TypeError:
                        body = body.encode('UTF-8')
                body = self._replace_response_values(body)

        response, content = http.request(
            full_url,
            method=method,
            headers=headers,
            body=body
        )

        # Set location attribute for follow on requests
        if 'location' in response:
            self.location = response['location']

        self._assert_response(response, content, test['status'],
                              headers=test['response_headers'],
                              expected=test['response_strings'],
                              json_paths=test['response_json_paths'])

    def _test_data_to_string(self, data):
        """Turn the request data into a string."""
        if isinstance(data, str):
            if data.startswith('<@'):
                info = self._load_data_file(data.replace('<@', '', 1))
                return info, True
            else:
                return data, False
        else:
            return json.dumps(data), False

    def _test_headers(self, headers, response):
        """Compare expected headers with actual headers."""
        for header in headers:
            header_value = headers[header].replace('$SCHEME', self.scheme)
            header_value = header_value.replace('$NETLOC', self.netloc)
            self.assertEqual(header_value, response[header],
                             'Expect header %s with value %s, got %s' %
                             (header, header_value, response[header]))

    def _test_status(self, expected_status, observed_status):
        """Confirm we got the expected status.

        If the status contains one or more || then it is treated as a
        list of acceptable statuses.
        """
        expected_status = str(expected_status)
        if '||' in expected_status:
            statii = [stat.strip() for stat in expected_status.split('||')]
        else:
            statii = [expected_status.strip()]
        self.assertIn(observed_status, statii)


class ServerError(Exception):
    """A catchall ServerError."""
    pass



# Copyright 2014, 2015 Red Hat
#
# Authors: Chris Dent <chdent@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Generate HTTP tests from YAML files

Each HTTP request is its own TestCase and can be requested to be run in
isolation from other tests. If it is member of a sequence of requests,
prior requests will be run.

A sequence is represented by an ordered list in a single YAML file.

Each sequence becomes a TestSuite.

An entire directory of YAML files is a TestSuite of TestSuites.
"""

import glob
import inspect
import os
from unittest import suite
import uuid

import httplib2
import wsgi_intercept
from wsgi_intercept import httplib2_intercept
import yaml

from .suite import GabbiSuite
from .case import HTTPTestCase


# Empty test from which all others inherit
BASE_TEST = {
    'name': '',
    'desc': '',
    'ssl': False,
    'redirects': False,
    'method': 'GET',
    'url': '',
    'status': '200',
    'request_headers': {},
    'response_headers': {},
    'response_strings': None,
    'response_json_paths': None,
    'data': '',
}


class TestBuilder(type):
    """Metaclass to munge a dynamically created test.
    """

    required_attributes = {'has_run': False}

    def __new__(mcs, name, bases, attributes):
        attributes.update(mcs.required_attributes)
        return type.__new__(mcs, name, bases, attributes)


def build_tests(path, loader, host=None, port=8001, intercept=None,
                test_file_name=None, fixture_module=None):
    """Read YAML files from a directory to create tests.

    Each YAML file represents an ordered sequence of HTTP requests.
    """
    top_suite = suite.TestSuite()
    http = httplib2.Http()

    if test_file_name is None:
        test_file_name = inspect.stack()[1]
        test_file_name = os.path.splitext(os.path.basename(
            test_file_name[1]))[0]

    # Return an empty suite if we have no host to access, either via
    # a real host or an intercept
    if not host and not intercept:
        return top_suite

    if intercept:
        host = install_intercept(intercept, port)

    test_directory = path
    path = '%s/*.yaml' % path

    key_test = set(BASE_TEST.keys())

    for test_file in glob.iglob(path):
        file_suite = GabbiSuite()
        test_yaml = load_yaml(test_file)
        test_data = test_yaml['tests']
        fixtures = test_yaml.get('fixtures', None)
        test_base_name = os.path.splitext(os.path.basename(test_file))[0]

        # Set defaults from BASE_TESTS the update those defaults
        # which any defaults set in the YAML file.
        base_test_data = dict(BASE_TEST)
        base_test_data.update(test_yaml.get('defaults', {}))

        fixture_classes = []
        if fixtures and fixture_module:
            for fixture_class in fixtures:
                fixture_classes.append(getattr(fixture_module, fixture_class))

        prior_test = None
        for test_datum in test_data:
            test = dict(base_test_data)
            test.update(test_datum)
            test_name = '%s_%s_%s' % (test_file_name,
                                      test_base_name,
                                      test['name'].lower().replace(' ', '_'))
            if set(test.keys()) != key_test:
                raise ValueError('Invalid Keys in test %s' % test_name)

            # Use metaclasses to build a class of the necessary type
            # with relevant arguments.
            klass = TestBuilder(test_name, (HTTPTestCase,),
                                {'test_data': test,
                                 'test_directory': test_directory,
                                 'fixtures': fixture_classes,
                                 'http': http,
                                 'host': host,
                                 'port': port,
                                 'prior': prior_test})

            tests = loader.loadTestsFromTestCase(klass)
            this_test = tests._tests[0]
            file_suite.addTest(this_test)
            prior_test = this_test

        top_suite.addTest(file_suite)

    return top_suite


def factory(wsgi_app):
    """Satisfy a bad API."""
    return wsgi_app


def install_intercept(wsgi_callable, port):
    """Install a wsgi-intercept on a random hostname."""
    hostname = str(uuid.uuid4())
    httplib2_intercept.install()
    wsgi_intercept.add_wsgi_intercept(hostname, port, factory(wsgi_callable))
    return hostname


def load_yaml(yaml_file):
    """Read and parse any YAML file. Let exceptions flow where they may."""
    with open(yaml_file) as source:
        return yaml.safe_load(source.read())



# Copyright 2014, 2015 Red Hat
#
# Authors: Chris Dent <chdent@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import suite

from . import fixture


class GabbiSuite(suite.TestSuite):
    """A TestSuite with fixtures.

    The suite wraps the tests with a set of nested context managers that
    operate as fixtures.
    """

    def run(self, result, debug=False):
        """Override TestSuite run to start suite-level fixtures.

        To avoid exception confusion, use a null Fixture when there
        are no fixtures.
        """

        # If there are fixtures, nest in their context.
        fixtures = [fixture.GabbiFixture]
        try:
            fixtures = self._tests[0].fixtures
        except AttributeError:
            pass
        with fixture.nest([fix() for fix in fixtures]):
            result = super(GabbiSuite, self).run(result, debug)

        return result



# Copyright 2014, 2015 Red Hat
#
# Authors: Chris Dent <chdent@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Manage fixtures for gabbi at the test suite level."""

import sys
from contextlib import contextmanager

import six


class GabbiFixtureError(Exception):
    """Generic exception for GabbiFixture."""
    pass


class GabbiFixture(object):
    """A context manager that operates as a fixture.

    Subclasses must implement start_fixture and stop_fixture, each of which
    contain the logic for stopping and starting whatever the fixture is.
    What a fixture is is left as an exercise for the implementor.

    These context managers will be nested so any actually work needs to
    happen in `start_fixture` and `stop_fixture` and not in ``__init__``.
    Otherwise exception handling will not work properly.
    """

    def __enter__(self):
        self.start_fixture()

    def __exit__(self, type, value, traceback):
        self.stop_fixture()

    def start_fixture(self):
        """Implement the actual workings of starting the fixture here."""
        pass

    def stop_fixture(self):
        """Implement the actual workings of stopping the fixture here."""
        pass


@contextmanager
def nest(fixtures):
    """Nest a series of fixtures.

    This is duplicated from ``nested`` in the stdlib, which has been
    deprecated because of issues with how exceptions are difficult to
    handle during ``__init__``. Gabbi needs to nest an unknown number
    of fixtures dynamically, so the ``with`` syntax that replaces
    ``nested`` will not work.
    """
    vars = []
    exits = []
    exc = (None, None, None)
    try:
        for fixture in fixtures:
            enter = fixture.__enter__
            exit = fixture.__exit__
            vars.append(enter())
            exits.append(exit)
        yield vars
    except:
        exc = sys.exc_info()
    finally:
        while exits:
            exit = exits.pop()
            try:
                if exit(*exc):
                    exc = (None, None, None)
            except:
                exc = sys.exc_info()
        if exc != (None, None, None):
            six.reraise(exc[0], exc[1], exc[2])



#
# Copyright 2014, 2015 Red Hat. All Rights Reserved.
#
# Author: Chris Dent <chdent@redhat.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A sample test module to exercise the code.

For the sake of exploratory development.
"""


import json
import os
import sys

from six.moves.urllib import parse as urlparse

from gabbi import driver
from gabbi import fixture


TESTS_DIR = 'gabbits_intercept'
METHODS = ['GET', 'PUT', 'POST', 'DELETE', 'PATCH']


class TestFixtureOne(fixture.GabbiFixture):
    """Drive the fixture testing weakly."""
    pass


class TestFixtureTwo(fixture.GabbiFixture):
    """Drive the fixture testing weakly."""
    pass


class SimpleWsgi(object):
    """A simple wsgi application to use in tests."""

    def __call__(self, environ, start_response):
        request_method = environ['REQUEST_METHOD'].upper()
        query_data = urlparse.parse_qs(environ.get('QUERY_STRING', ''))

        request_url = environ.get('REQUEST_URI',
                                  environ.get('RAW_URI', 'unknown'))

        path, query, fragment = urlparse.urlsplit(request_url)[2:]
        server_name = environ.get('SERVER_NAME')
        server_port = environ.get('SERVER_PORT')
        server_scheme = environ.get('wsgi.url_scheme')
        if server_port not in ['80', '443']:
            netloc = '%s:%s' % (server_name, server_port)
        else:
            netloc = server_name

        request_url = urlparse.urlunsplit((server_scheme, netloc, path,
                                           query, fragment))

        accept_header = environ.get('HTTP_ACCEPT')

        if accept_header:
            content_type = accept_header
        else:
            content_type = 'application/json'

        headers = [
            ('X-Gabbi-method', request_method),
            ('Content-Type', content_type),
            ('X-Gabbi-url', request_url),
        ]

        if request_method not in METHODS:
            headers.append(
                ('Allow', ', '.join(METHODS)))
            start_response('405 Method Not Allowed', headers)
            return []

        if request_method.startswith('P'):
            body = environ['wsgi.input'].read()
            if body:
                if environ.get('CONTENT_TYPE', '') == 'application/json':
                    body_data = json.loads(body.decode('utf-8'))
                    if query_data:
                        query_data.update(body_data)
                    else:
                        query_data = body_data
            headers.append(('Location', request_url))

        start_response('200 OK', headers)

        query_output = json.dumps(query_data)
        return [query_output.encode('utf-8')]


def load_tests(loader, tests, pattern):
    """Provide a TestSuite to the discovery process."""
    test_dir = os.path.join(os.path.dirname(__file__), TESTS_DIR)
    return driver.build_tests(test_dir, loader, host=None,
                              intercept=SimpleWsgi,
                              fixture_module=sys.modules[__name__])
